# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import csv
import json
import logging
import math
import os
import re
import subprocess
import textwrap
import uuid
from typing import Any, Dict, List, Optional, Tuple, Union

from toil.fileStores.abstractFileStore import AbstractFileStore
from toil.lib.conversions import bytes_in_unit
from toil.lib.resources import glob  # type: ignore
from toil.wdl.wdl_types import WDLFile, WDLPair

logger = logging.getLogger(__name__)


class WDLRuntimeError(Exception):
    """ WDL-related run-time error."""

    def __init__(self, message):
        super().__init__(message)


class WDLJSONEncoder(json.JSONEncoder):
    """
    Extended JSONEncoder to support WDL-specific JSON encoding.
    """

    def default(self, obj):
        if isinstance(obj, WDLPair):
            return obj.to_dict()
        return json.JSONEncoder.default(self, obj)


def generate_docker_bashscript_file(temp_dir, docker_dir, globs, cmd, job_name):
    '''
    Creates a bashscript to inject into a docker container for the job.

    This script wraps the job command(s) given in a bash script, hard links the
    outputs and returns an "rc" file containing the exit code.  All of this is
    done in an effort to parallel the Broad's cromwell engine, which is the
    native WDL runner.  As they've chosen to write and then run a bashscript for
    every command, so shall we.

    :param temp_dir: The current directory outside of docker to deposit the
                     bashscript into, which will be the bind mount that docker
                     loads files from into its own containerized filesystem.
                     This is usually the tempDir created by this individual job
                     using 'tempDir = job.fileStore.getLocalTempDir()'.
    :param docker_dir: The working directory inside of the docker container
                       which is bind mounted to 'temp_dir'.  By default this is
                       'data'.
    :param globs: A list of expected output files to retrieve as glob patterns
                  that will be returned as hard links to the current working
                  directory.
    :param cmd: A bash command to be written into the bash script and run.
    :param job_name: The job's name, only used to write in a file name
                     identifying the script as written for that job.
                     Will be used to call the script later.
    :return: Nothing, but it writes and deposits a bash script in temp_dir
             intended to be run inside of a docker container for this job.
    '''
    wdl_copyright = heredoc_wdl('''        \n
        # Borrowed/rewritten from the Broad's Cromwell implementation.  As
        # that is under a BSD-ish license, I include here the license off
        # of their GitHub repo.  Thank you Broadies!

        # Copyright (c) 2015, Broad Institute, Inc.
        # All rights reserved.

        # Redistribution and use in source and binary forms, with or without
        # modification, are permitted provided that the following conditions are met:

        # * Redistributions of source code must retain the above copyright notice, this
        #   list of conditions and the following disclaimer.

        # * Redistributions in binary form must reproduce the above copyright notice,
        #   this list of conditions and the following disclaimer in the documentation
        #   and/or other materials provided with the distribution.

        # * Neither the name Broad Institute, Inc. nor the names of its
        #   contributors may be used to endorse or promote products derived from
        #   this software without specific prior written permission.

        # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
        # AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
        # IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
        # DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
        # FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
        # DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
        # SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
        # CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
        # OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
        # OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE

        # make a temp directory w/identifier
        ''')
    prefix_dict = {"docker_dir": docker_dir,
                   "cmd": cmd}
    bashfile_prefix = heredoc_wdl('''
        tmpDir=$(mktemp -d /{docker_dir}/execution/tmp.XXXXXX)
        chmod 777 $tmpDir
        # set destination for java to deposit all of its files
        export _JAVA_OPTIONS=-Djava.io.tmpdir=$tmpDir
        export TMPDIR=$tmpDir

        (
        cd /{docker_dir}/execution
        {cmd}
        )

        # gather the input command return code
        echo $? > "$tmpDir/rc.tmp"

        ''', prefix_dict)

    bashfile_string = '#!/bin/bash' + wdl_copyright + bashfile_prefix

    begin_globbing_string = heredoc_wdl('''
        (
        mkdir "$tmpDir/globs"
        ''')

    bashfile_string = bashfile_string + begin_globbing_string

    for glob_input in globs:
        add_this_glob = \
            '( ln -L ' + glob_input + \
            ' "$tmpDir/globs" 2> /dev/null ) || ( ln ' + glob_input + \
            ' "$tmpDir/globs" )\n'
        bashfile_string = bashfile_string + add_this_glob

    bashfile_suffix = heredoc_wdl('''
        )

        # flush RAM to disk
        sync

        mv "$tmpDir/rc.tmp" "$tmpDir/rc"
        chmod -R 777 $tmpDir
        ''')

    bashfile_string = bashfile_string + bashfile_suffix

    with open(os.path.join(temp_dir, job_name + '_script.sh'), 'w') as bashfile:
        bashfile.write(bashfile_string)


def process_single_infile(wdl_file: WDLFile, fileStore: AbstractFileStore) -> WDLFile:
    f = wdl_file.file_path
    logger.info(f'Importing {f} into the jobstore.')
    if f.startswith('http://') or f.startswith('https://') or \
            f.startswith('file://') or f.startswith('wasb://'):
        filepath = fileStore.importFile(f)
        preserveThisFilename = os.path.basename(f)
    elif f.startswith('s3://'):
        try:
            filepath = fileStore.importFile(f)
            preserveThisFilename = os.path.basename(f)
        except:
            from toil.lib.ec2nodes import EC2Regions
            success = False
            for region in EC2Regions:
                try:
                    html_path = f'http://s3.{region}.amazonaws.com/' + f[5:]
                    filepath = fileStore.importFile(html_path)
                    preserveThisFilename = os.path.basename(f)
                    success = True
                except:
                    pass
            if not success:
                raise RuntimeError('Unable to import: ' + f)
    elif f.startswith('gs://'):
        f = 'https://storage.googleapis.com/' + f[5:]
        filepath = fileStore.importFile(f)
        preserveThisFilename = os.path.basename(f)
    else:
        filepath = fileStore.importFile("file://" + os.path.abspath(f))
        preserveThisFilename = os.path.basename(f)
    return WDLFile(file_path=filepath, file_name=preserveThisFilename, imported=True)


def process_infile(f: Any, fileStore: AbstractFileStore):
    """
    Takes any input and imports the WDLFile into the fileStore.

    This returns the input importing all WDLFile instances to the fileStore.  Toil
    does not preserve a file's original name upon import and so the WDLFile also keeps
    track of this.

    :param f: A primitive, WDLFile, or a container. A file needs to be a WDLFile instance
              to be imported.
    :param fileStore: The fileStore object that is called to load files into the fileStore.
    """
    if isinstance(f, WDLFile):
        # check if this has already been imported into the fileStore
        if f.imported:
            return f
        else:
            return process_single_infile(f, fileStore)
    elif isinstance(f, list):
        # recursively call process_infile() to handle cases like Array[Map[String, File]]
        return [process_infile(sf, fileStore) for sf in f]
    elif isinstance(f, WDLPair):
        f.left = process_infile(f.left, fileStore)
        f.right = process_infile(f.right, fileStore)
        return f
    elif isinstance(f, dict):
        return {process_infile(k, fileStore): process_infile(v, fileStore) for k, v in f.items()}
    elif isinstance(f, (int, str, bool, float)):
        return f
    else:
        raise WDLRuntimeError(f'Error processing file: {str(f)}')


def sub(input_str: str, pattern: str, replace: str) -> str:
    """
    Given 3 String parameters `input`, `pattern`, `replace`, this function will
    replace any occurrence matching `pattern` in `input` by `replace`.
    `pattern` is expected to be a regular expression. Details of regex evaluation
    will depend on the execution engine running the WDL.

    WDL syntax: String sub(String, String, String)
    """

    if isinstance(input_str, WDLFile):
        input_str = input_str.file_name
    if isinstance(pattern, WDLFile):
        pattern = pattern.file_name
    if isinstance(replace, WDLFile):
        replace = replace.file_name

    return re.sub(pattern=str(pattern), repl=str(replace), string=str(input_str))


def defined(i):
    if i:
        return True
    return False


def process_single_outfile(wdl_file: WDLFile, fileStore, workDir, outDir) -> WDLFile:
    f = wdl_file.file_path
    if os.path.exists(f):
        output_f_path = f
    elif os.path.exists(os.path.abspath(f)):
        output_f_path = os.path.abspath(f)
    elif os.path.exists(os.path.join(workDir, 'execution', f)):
        output_f_path = os.path.join(workDir, 'execution', f)
    elif os.path.exists(os.path.join('execution', f)):
        output_f_path = os.path.join('execution', f)
    elif os.path.exists(os.path.join(workDir, f)):
        output_f_path = os.path.join(workDir, f)
    elif os.path.exists(os.path.join(outDir, f)):
        output_f_path = os.path.join(outDir, f)
    else:
        tmp = subprocess.check_output(['ls', '-lha', workDir]).decode('utf-8')
        exe = subprocess.check_output(['ls', '-lha', os.path.join(workDir, 'execution')]).decode('utf-8')
        for std_file in ('stdout', 'stderr'):
            std_file = os.path.join(workDir, 'execution', std_file)
            if os.path.exists(std_file):
                with open(std_file, 'rb') as f:
                    logger.info(f.read())

        raise RuntimeError('OUTPUT FILE: {} was not found in {}!\n'
                           '{}\n\n'
                           '{}\n'.format(f, os.getcwd(), tmp, exe))
    output_file = fileStore.writeGlobalFile(output_f_path)
    preserveThisFilename = os.path.basename(output_f_path)
    fileStore.export_file(output_file, "file://" + os.path.join(os.path.abspath(outDir), preserveThisFilename))
    return WDLFile(file_path=output_file, file_name=preserveThisFilename, imported=True)


def process_outfile(f, fileStore, workDir, outDir):
    if isinstance(f, WDLFile):
        return process_single_outfile(f, fileStore, workDir, outDir)
    elif isinstance(f, list):
        # recursively call process_outfile() to handle cases like Array[Map[String, File]]
        return [process_outfile(sf, fileStore, workDir, outDir) for sf in f]
    elif isinstance(f, WDLPair):
        f.left = process_outfile(f.left, fileStore, workDir, outDir)
        f.right = process_outfile(f.right, fileStore, workDir, outDir)
        return f
    elif isinstance(f, dict):
        return {process_outfile(k, fileStore, workDir, outDir):
                process_outfile(v, fileStore, workDir, outDir) for k, v in f.items()}
    elif isinstance(f, (int, str, bool, float)):
        return f
    else:
        raise WDLRuntimeError(f'Error processing file: {str(f)}')


def abspath_single_file(f: WDLFile, cwd: str) -> WDLFile:
    path = f.file_path
    if path != os.path.abspath(path):
        f.file_path = os.path.join(cwd, path)
    return f


def abspath_file(f: Any, cwd: str):
    if not f:
        # in the case of "optional" files (same treatment in 'process_and_read_file()')
        # TODO: handle this at compile time, not here
        return ''
    if isinstance(f, WDLFile):
        # check if this has already been imported into the fileStore
        if f.imported:
            return f
        path = f.file_path
        if path.startswith('s3://') or path.startswith('http://') or path.startswith('https://') or \
                path.startswith('file://') or path.startswith('wasb://') or path.startswith('gs://'):
            return f
        return abspath_single_file(f, cwd)
    elif isinstance(f, list):
        # recursively call abspath_file() to handle cases like Array[Map[String, File]]
        return [abspath_file(sf, cwd) for sf in f]
    elif isinstance(f, WDLPair):
        f.left = abspath_file(f.left, cwd)
        f.right = abspath_file(f.right, cwd)
        return f
    elif isinstance(f, dict):
        return {abspath_file(k, cwd): abspath_file(v, cwd) for k, v in f.items()}
    elif isinstance(f, (int, str, bool, float)):
        return f
    else:
        raise WDLRuntimeError(f'Error processing file: ({str(f)}) of type: ({str(type(f))}).')


def read_single_file(f: WDLFile, tempDir, fileStore, docker=False) -> str:
    import os
    try:
        fpath = fileStore.readGlobalFile(f.file_path, userPath=os.path.join(tempDir, f.file_name))
    except:
        fpath = os.path.join(tempDir, f.file_name)
    return fpath


def read_file(f: Any, tempDir: str, fileStore: AbstractFileStore, docker: bool = False):
    if isinstance(f, WDLFile):
        return read_single_file(f, tempDir, fileStore, docker=docker)
    elif isinstance(f, list):
        # recursively call read_file() to handle cases like Array[Map[String, File]]
        return [read_file(sf, tempDir, fileStore, docker=docker) for sf in f]
    elif isinstance(f, WDLPair):
        f.left = read_file(f.left, tempDir, fileStore, docker=docker)
        f.right = read_file(f.right, tempDir, fileStore, docker=docker)
        return f
    elif isinstance(f, dict):
        return {read_file(k, tempDir, fileStore, docker=docker):
                read_file(v, tempDir, fileStore, docker=docker) for k, v in f.items()}
    elif isinstance(f, (int, str, bool, float)):
        return f
    else:
        raise WDLRuntimeError(f'Error processing file: {str(f)}')


def process_and_read_file(f, tempDir, fileStore, docker=False):
    if not f:
        # in the case of "optional" files (same treatment in 'abspath_file()')
        # TODO: handle this at compile time, not here and change to the empty string
        return None
    processed_file = process_infile(f, fileStore)
    return read_file(processed_file, tempDir, fileStore, docker=docker)


def generate_stdout_file(output, tempDir, fileStore, stderr=False):
    """
    Create a stdout (or stderr) file from a string or bytes object.

    :param str|bytes output: A str or bytes object that holds the stdout/stderr text.
    :param str tempDir: The directory to write the stdout file.
    :param fileStore: A fileStore object.
    :param bool stderr: If True, a stderr instead of a stdout file is generated.
    :return: The file path to the generated file.
    """
    if output is None:
        # write an empty file if there's no stdout/stderr.
        output = b''
    elif isinstance(output, str):
        output = bytes(output, encoding='utf-8')

    # TODO: we need a way to differentiate the stdout/stderr files in the workflow after execution.
    # Cromwell generates a folder for each task so the file is simply named stdout and lives in
    # the task execution folder. This is not the case with Toil. Though, this would not be a
    # problem with intermediate stdout files as each task has its own temp folder.
    name = 'stderr' if stderr else 'stdout'
    local_path = os.path.join(tempDir, 'execution', name)

    # import to fileStore then read to local temp file
    with fileStore.writeGlobalFileStream(cleanup=True, basename=name) as (stream, file_id):
        stream.write(output)

    assert file_id is not None
    return fileStore.readGlobalFile(fileStoreID=file_id, userPath=local_path)


def parse_memory(memory):
    """
    Parses a string representing memory and returns
    an integer # of bytes.

    :param memory:
    :return:
    """
    memory = str(memory)
    if 'None' in memory:
        return 2147483648  # toil's default
    try:
        import re
        raw_mem_split = re.split('([a-zA-Z]+)', memory)
        mem_split = []

        for r in raw_mem_split:
            if r:
                mem_split.append(r.replace(' ', ''))

        if len(mem_split) == 1:
            return int(memory)

        if len(mem_split) == 2:
            num = mem_split[0]
            unit = mem_split[1]
            return int(float(num) * bytes_in_unit(unit))
        else:
            raise RuntimeError(f'Memory parsing failed: {memory}')
    except:
        return 2147483648  # toil's default


def parse_cores(cores):
    cores = str(cores)
    if 'None' in cores:
        return 1  # toil's default
    if cores:
        return float(cores)
    else:
        return 1


def parse_disk(disk):
    disk = str(disk)
    if 'None' in disk:
        return 2147483648  # toil's default
    try:
        total_disk = 0
        disks = disk.split(',')
        for d in disks:
            d = d.strip().split(' ')
            if len(d) > 1:
                for part in d:
                    if is_number(part):
                        total_disk += parse_memory(f'{part} GB')
            else:
                return parse_memory(d[0]) if parse_memory(d[0]) > 2147483648 else 2147483648
        return total_disk if total_disk > 2147483648 else 2147483648
    except:
        return 2147483648  # toil's default


def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def size(f: Optional[Union[str, WDLFile, List[Union[str, WDLFile]]]] = None,
         unit: Optional[str] = 'B',
         fileStore: Optional[AbstractFileStore] = None) -> float:
    """
    Given a `File` and a `String` (optional), returns the size of the file in Bytes
    or in the unit specified by the second argument.

    Supported units are KiloByte ("K", "KB"), MegaByte ("M", "MB"), GigaByte
    ("G", "GB"), TeraByte ("T", "TB") (powers of 1000) as well as their binary version
    (https://en.wikipedia.org/wiki/Binary_prefix) "Ki" ("KiB"), "Mi" ("MiB"),
    "Gi" ("GiB"), "Ti" ("TiB") (powers of 1024). Default unit is Bytes ("B").

    WDL syntax: Float size(File, [String])
    Varieties:  Float size(File?, [String])
                Float size(Array[File], [String])
                Float size(Array[File?], [String])
    """

    if f is None:
        return 0

    # it is possible that size() is called directly (e.g.: size('file')) and so it is not treated as a file.
    if isinstance(f, str):
        f = WDLFile(file_path=f)
    elif isinstance(f, list):
        f = [WDLFile(file_path=sf) if isinstance(sf, str) else sf for sf in f]

    assert isinstance(f, (WDLFile, list)), f'size() excepts a "File" or "File?" argument!  Not: {type(f)}'

    # validate the input. fileStore is only required if the input is not processed.
    f = process_infile(f, fileStore)

    divisor = bytes_in_unit(unit)

    if isinstance(f, list):
        total_size = sum(file.file_path.size for file in f)
        return total_size / divisor

    fileID = f.file_path
    return fileID.size / divisor


def select_first(values):
    for var in values:
        if var:
            return var
    raise ValueError(f'No defined variables found for select_first array: {str(values)}')


def combine_dicts(dict1, dict2):
    combineddict= {}
    for k, v in dict1.items():
        counter1 = 0
        while isinstance(v, list):
            counter1 += 1
            v = v[0]
        break

    for k, v in dict2.items():
        counter2 = 0
        while isinstance(v, list):
            counter2 += 1
            v = v[0]
        break

    for k in dict1:
        if counter1 > counter2:
            combineddict[k] = dict1[k]
            combineddict[k].append(dict2[k])
        elif counter1 < counter2:
            combineddict[k] = dict2[k]
            combineddict[k].append(dict1[k])
        else:
            combineddict[k] = [dict1[k], dict2[k]]
    return combineddict


def basename(path, suffix=None):
    """https://software.broadinstitute.org/wdl/documentation/article?id=10554"""
    path = path.strip()
    if suffix:
        suffix = suffix.strip()
        if path.endswith(suffix):
            path = path[:-len(suffix)]
    return os.path.basename(path)


def heredoc_wdl(template, dictionary={}, indent=''):
    template = textwrap.dedent(template).format(**dictionary)
    return template.replace('\n', '\n' + indent) + '\n'


def floor(i: Union[int, float]) -> int:
    """
    Converts a Float value into an Int by rounding down to the next lower integer.
    """
    return math.floor(i)


def ceil(i: Union[int, float]) -> int:
    """
    Converts a Float value into an Int by rounding up to the next higher integer.
    """
    return math.ceil(i)


def read_lines(path: str) -> List[str]:
    """
    Given a file-like object (`String`, `File`) as a parameter, this will read each
    line as a string and return an `Array[String]` representation of the lines in
    the file.

    WDL syntax: Array[String] read_lines(String|File)
    """
    # file should already be imported locally via `process_and_read_file`
    with open(path) as f:
        return f.read().rstrip('\n').split('\n')


def read_tsv(path: str, delimiter: str = '\t') -> List[List[str]]:
    """
    Take a tsv filepath and return an array; e.g. [[],[],[]].

    For example, a file containing:

    1   2   3
    4   5   6
    7   8   9

    would return the array: [['1','2','3'], ['4','5','6'], ['7','8','9']]

    WDL syntax: Array[Array[String]] read_tsv(String|File)
    """
    tsv_array = []
    with open(path) as f:
        data_file = csv.reader(f, delimiter=delimiter)
        for line in data_file:
            tsv_array.append(line)
    return tsv_array


def read_csv(path: str) -> List[List[str]]:
    """
    Take a csv filepath and return an array; e.g. [[],[],[]].

    For example, a file containing:

    1,2,3
    4,5,6
    7,8,9

    would return the array: [['1','2','3'], ['4','5','6'], ['7','8','9']]
    """
    return read_tsv(path, delimiter=",")


def read_json(path: str) -> Any:
    """
    The `read_json()` function takes one parameter, which is a file-like object
     (`String`, `File`) and returns a data type which matches the data
     structure in the JSON file. See
    https://github.com/openwdl/wdl/blob/main/versions/development/SPEC.md#mixed-read_jsonstringfile

    WDL syntax: mixed read_json(String|File)
    """
    with open(path) as f:
        return json.load(f)


def read_map(path: str) -> Dict[str, str]:
    """
    Given a file-like object (`String`, `File`) as a parameter, this will read each
    line from a file and expect the line to have the format `col1\tcol2`. In other
    words, the file-like object must be a two-column TSV file.

    WDL syntax: Map[String, String] read_map(String|File)
    """
    d = dict()
    with open(path) as f:
        for line in f:
            line = line.rstrip()
            if not line:
                # remove extra lines
                continue
            key, value = line.split('\t', 1)
            d[key] = value.strip()
    return d


def read_int(path: Union[str, WDLFile]) -> int:
    """
    The `read_int()` function takes a file path which is expected to contain 1
    line with 1 integer on it. This function returns that integer.

    WDL syntax: Int read_int(String|File)
    """
    if isinstance(path, WDLFile):
        path = path.file_path

    with open(path) as f:
        return int(f.read().strip())


def read_string(path: Union[str, WDLFile]) -> str:
    """
    The `read_string()` function takes a file path which is expected to contain 1
    line with 1 string on it. This function returns that string.

    WDL syntax: String read_string(String|File)
    """
    if isinstance(path, WDLFile):
        path = path.file_path

    with open(path) as f:
        return str(f.read().strip())


def read_float(path: Union[str, WDLFile]) -> float:
    """
    The `read_float()` function takes a file path which is expected to contain 1
    line with 1 floating point number on it. This function returns that float.

    WDL syntax: Float read_float(String|File)
    """
    if isinstance(path, WDLFile):
        path = path.file_path

    with open(path) as f:
        return float(f.read().strip())


def read_boolean(path: Union[str, WDLFile]) -> bool:
    """
    The `read_boolean()` function takes a file path which is expected to contain 1
    line with 1 Boolean value (either "true" or "false" on it). This function
    returns that Boolean value.

    WDL syntax: Boolean read_boolean(String|File)
    """
    if isinstance(path, WDLFile):
        path = path.file_path

    with open(path) as f:
        return f.read().strip().lower() == 'true'


def _get_temp_file_path(function_name: str, temp_dir: Optional[str] = None) -> str:
    """
    Get a unique path with basename in the format of "{function_name}_{UUID}.tmp".
    """

    if not temp_dir:
        temp_dir = os.getcwd()

    # Cromwell uses the MD5 checksum of the content as part of the file name. We use a UUID instead
    # for now, since we're writing line by line via a context manager.
    # md5sum = hashlib.md5(content).hexdigest()
    # name = f'{function_name}_{md5sum}.tmp'

    name = f'{function_name}_{uuid.uuid4()}.tmp'
    return os.path.join(temp_dir, 'execution', name)


def write_lines(in_lines: List[str],
                temp_dir: Optional[str] = None,
                file_store: Optional[AbstractFileStore] = None) -> str:
    """
    Given something that's compatible with `Array[String]`, this writes each element
    to it's own line on a file.  with newline `\n` characters as line separators.

    WDL syntax: File write_lines(Array[String])
    """
    assert isinstance(in_lines, list), f'write_lines() requires "{in_lines}" to be a list!  Not: {type(in_lines)}'

    path = _get_temp_file_path('write_lines', temp_dir)

    with open(path, 'w') as file:
        for line in in_lines:
            file.write(f'{line}\n')

    if file_store:
        file_store.writeGlobalFile(path, cleanup=True)

    return path


def write_tsv(in_tsv: List[List[str]],
              delimiter: str = '\t',
              temp_dir: Optional[str] = None,
              file_store: Optional[AbstractFileStore] = None) -> str:
    """
    Given something that's compatible with `Array[Array[String]]`, this writes a TSV
    file of the data structure.

    WDL syntax: File write_tsv(Array[Array[String]])
    """
    assert isinstance(in_tsv, list), f'write_tsv() requires "{in_tsv}" to be a list!  Not: {type(in_tsv)}'

    path = _get_temp_file_path('write_tsv', temp_dir)

    with open(path, 'w') as file:
        tsv_writer = csv.writer(file, delimiter=delimiter)
        for row in in_tsv:
            tsv_writer.writerow(row)

    if file_store:
        file_store.writeGlobalFile(path, cleanup=True)

    return path


def write_json(in_json: Any,
               indent: Union[None, int, str] = None,
               separators: Optional[Tuple[str, str]] = (',', ':'),
               temp_dir: Optional[str] = None,
               file_store: Optional[AbstractFileStore] = None) -> str:
    """
    Given something with any type, this writes the JSON equivalent to a file. See
    the table in the definition of
    https://github.com/openwdl/wdl/blob/main/versions/development/SPEC.md#mixed-read_jsonstringfile

    WDL syntax: File write_json(mixed)
    """

    path = _get_temp_file_path('write_json', temp_dir)

    with open(path, 'w') as file:
        file.write(json.dumps(in_json, indent=indent, separators=separators, cls=WDLJSONEncoder))

    if file_store:
        file_store.writeGlobalFile(path, cleanup=True)

    return path


def write_map(in_map: Dict[str, str],
              temp_dir: Optional[str] = None,
              file_store: Optional[AbstractFileStore] = None) -> str:
    """
    Given something that's compatible with `Map[String, String]`, this writes a TSV
     file of the data structure.

    WDL syntax: File write_map(Map[String, String])
    """
    assert isinstance(in_map, dict), f'write_map() requires "{in_map}" to be a dict!  Not: {type(in_map)}'

    path = _get_temp_file_path('write_map', temp_dir)

    with open(path, 'w') as file:
        for key, val in in_map.items():
            file.write(f'{key}\t{val}\n')

    if file_store:
        file_store.writeGlobalFile(path, cleanup=True)

    return path


def wdl_range(num: int) -> List[int]:
    """
    Given an integer argument, the range function creates an array of integers of
    length equal to the given argument.

    WDL syntax: Array[Int] range(Int)
    """
    if not (isinstance(num, int) and num >= 0):
        raise WDLRuntimeError(f'range() requires an integer greater than or equal to 0 (but got {num})')

    return list(range(num))


def transpose(in_array: List[List[Any]]) -> List[List[Any]]:
    """
    Given a two dimensional array argument, the transpose function transposes the
    two dimensional array according to the standard matrix transpose rules.

    WDL syntax: Array[Array[X]] transpose(Array[Array[X]])
    """
    assert isinstance(in_array, list), f'transpose() requires "{in_array}" to be a list!  Not: {type(in_array)}'

    for arr in in_array:
        assert isinstance(arr, list), f'transpose() requires all collections to be a list!  Not: {type(arr)}'
        # zip() can handle this but Cromwell can not.
        assert len(arr) == len(in_array[0]), 'transpose() requires all collections have the same size!'

    return [list(i) for i in zip(*in_array)]


def length(in_array: List[Any]) -> int:
    """
    Given an Array, the `length` function returns the number of elements in the Array
    as an Int.
    """
    if not isinstance(in_array, list):
        # Cromwell throws an exception for anything other than a WDL Array
        raise WDLRuntimeError(f'length() requires ${in_array} to be a list!  Not: {type(in_array)}')

    return len(in_array)


def wdl_zip(left: List[Any], right: List[Any]) -> List[WDLPair]:
    """
    Return the dot product of the two arrays. If the arrays have different lengths
    it is an error.

    WDL syntax: Array[Pair[X,Y]] zip(Array[X], Array[Y])
    """
    if not isinstance(left, list) or not isinstance(right, list):
        raise WDLRuntimeError(f'zip() requires both inputs to be lists!  Not: {type(left)} and {type(right)}')

    if len(left) != len(right):
        raise WDLRuntimeError('zip() requires that input values have the same size!')

    return list(WDLPair(left=left_val, right=right_val) for left_val, right_val in zip(left, right))


def cross(left: List[Any], right: List[Any]) -> List[WDLPair]:
    """
    Return the cross product of the two arrays. Array[Y][1] appears before
    Array[X][1] in the output.

    WDL syntax: Array[Pair[X,Y]] cross(Array[X], Array[Y])
    """
    if not isinstance(left, list) or not isinstance(right, list):
        raise WDLRuntimeError(f'cross() requires both inputs to be Array[]!  Not: {type(left)} and {type(right)}')

    return list(WDLPair(left=left_val, right=right_val) for left_val in left for right_val in right)


def as_pairs(in_map: dict) -> List[WDLPair]:
    """
    Given a Map, the `as_pairs` function returns an Array containing each element
    in the form of a Pair. The key will be the left element of the Pair and the
    value the right element. The order of the the Pairs in the resulting Array
    is the same as the order of the key/value pairs in the Map.

    WDL syntax: Array[Pair[X,Y]] as_pairs(Map[X,Y])
    """
    if not isinstance(in_map, dict):
        raise WDLRuntimeError(f'as_pairs() requires "{in_map}" to be Map[]!  Not: {type(in_map)}')

    return list(WDLPair(left=k, right=v) for k, v in in_map.items())


def as_map(in_array: List[WDLPair]) -> dict:
    """
    Given an Array consisting of Pairs, the `as_map` function returns a Map in
    which the left elements of the Pairs are the keys and the right elements the
    values.

    WDL syntax: Map[X,Y] as_map(Array[Pair[X,Y]])
    """
    if not isinstance(in_array, list):
        raise WDLRuntimeError(f'as_map() requires "{in_array}" to be a list!  Not: {type(in_array)}')

    map = {}

    for pair in in_array:
        if map.get(pair.left):
            raise WDLRuntimeError('Cannot evaluate "as_map()" with duplicated keys.')

        map[pair.left] = pair.right

    return map


def keys(in_map: dict) -> list:
    """
    Given a Map, the `keys` function returns an Array consisting of the keys in
    the Map. The order of the keys in the resulting Array is the same as the
    order of the Pairs in the Map.

    WDL syntax: Array[X] keys(Map[X,Y])
    """

    return list(in_map.keys())


def collect_by_key(in_array: List[WDLPair]) -> dict:
    """
    Given an Array consisting of Pairs, the `collect_by_key` function returns a Map
    in which the left elements of the Pairs are the keys and the right elements the
    values.

    WDL syntax: Map[X,Array[Y]] collect_by_key(Array[Pair[X,Y]])
    """
    if not isinstance(in_array, list):
        raise WDLRuntimeError(f'as_map() requires "{in_array}" to be a list!  Not: {type(in_array)}')

    map = {}

    for pair in in_array:
        map.setdefault(pair.left, []).append(pair.right)

    return map


def flatten(in_array: List[list]) -> list:
    """
    Given an array of arrays, the `flatten` function concatenates all the member
    arrays in the order to appearance to give the result. It does not deduplicate
    the elements.

    WDL syntax: Array[X] flatten(Array[Array[X]])
    """
    assert isinstance(in_array, list), f'flatten() requires "{in_array}" to be a list!  Not: {type(in_array)}'

    arr = []

    for element in in_array:
        assert isinstance(element, list), f'flatten() requires all collections to be a list!  Not: {type(element)}'
        arr.extend(element)

    return arr



# Copyright (C) 2015-2022 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import inspect
import logging
import os
import re
import socket
import threading
from typing import (Any,
                    Callable,
                    Dict,
                    Iterable,
                    List,
                    Optional,
                    Tuple,
                    TypeVar,
                    Union,
                    cast)
from urllib.error import URLError
from urllib.request import urlopen

import boto3
import boto3.resources.base
import boto.connection
import botocore
from boto3 import Session
from botocore.client import Config
from botocore.credentials import JSONFileCache
from botocore.session import get_session

logger = logging.getLogger(__name__)

# A note on thread safety:
#
# Boto3 Session: Not thread safe, 1 per thread is required.
#
# Boto3 Resources: Not thread safe, one per thread is required.
#
# Boto3 Client: Thread safe after initialization, but initialization is *not*
# thread safe and only one can be being made at a time. They also are
# restricted to a single Python *process*.
#
# See: <https://stackoverflow.com/questions/52820971/is-boto3-client-thread-safe>

# We use this lock to control initialization so only one thread can be
# initializing Boto3 (or Boto2) things at a time.
_init_lock = threading.RLock()

def _new_boto3_session(region_name: Optional[str] = None) -> Session:
    """
    This is the One True Place where new Boto3 sessions should be made, and
    prepares them with the necessary credential caching. Does *not* cache
    sessions, because each thread needs its own caching.

    :param region_name: If given, the session will be associated with the given AWS region.
    """

    # Make sure to use credential caching when talking to Amazon via boto3
    # See https://github.com/boto/botocore/pull/1338/
    # And https://github.com/boto/botocore/commit/2dae76f52ae63db3304b5933730ea5efaaaf2bfc

    with _init_lock:
        botocore_session = get_session()
        botocore_session.get_component('credential_provider').get_provider(
            'assume-role').cache = JSONFileCache()

        return Session(botocore_session=botocore_session, region_name=region_name, profile_name=os.environ.get("TOIL_AWS_PROFILE", None))

class AWSConnectionManager:
    """
    Class that represents a connection to AWS. Caches Boto 3 and Boto 2 objects
    by region.

    Access to any kind of item goes through the particular method for the thing
    you want (session, resource, service, Boto2 Context), and then you pass the
    region you want to work in, and possibly the type of thing you want, as arguments.

    This class is intended to eventually enable multi-region clusters, where
    connections to multiple regions may need to be managed in the same
    provisioner.

    We also support None for a region, in which case no region will be
    passed to Boto/Boto3. The caller is responsible for implementing e.g.
    TOIL_AWS_REGION support.

    Since connection objects may not be thread safe (see
    <https://boto3.amazonaws.com/v1/documentation/api/1.14.31/guide/session.html#multithreading-or-multiprocessing-with-sessions>),
    one is created for each thread that calls the relevant lookup method.
    """

    # TODO: mypy is going to have !!FUN!! with this API because the final type
    # we get out (and whether it has the right methods for where we want to use
    # it) depends on having the right string value for the service. We could
    # also individually wrap every service we use, but that seems like a good
    # way to generate a lot of boring code.

    def __init__(self) -> None:
        """
        Make a new empty AWSConnectionManager.
        """
        # This stores Boto3 sessions in .item of a thread-local storage, by
        # region.
        self.sessions_by_region: Dict[Optional[str], threading.local] = collections.defaultdict(threading.local)
        # This stores Boto3 resources in .item of a thread-local storage, by
        # (region, service name, endpoint URL) tuples
        self.resource_cache: Dict[Tuple[Optional[str], str, Optional[str]], threading.local] = collections.defaultdict(threading.local)
        # This stores Boto3 clients in .item of a thread-local storage, by
        # (region, service name, endpoint URL) tuples
        self.client_cache: Dict[Tuple[Optional[str], str, Optional[str]], threading.local] = collections.defaultdict(threading.local)
        # This stores Boto 2 connections in .item of a thread-local storage, by
        # (region, service name) tuples.
        self.boto2_cache: Dict[Tuple[Optional[str], str], threading.local] = collections.defaultdict(threading.local)

    def session(self, region: Optional[str]) -> boto3.session.Session:
        """
        Get the Boto3 Session to use for the given region.
        """
        storage = self.sessions_by_region[region]
        if not hasattr(storage, 'item'):
            # This is the first time this thread wants to talk to this region
            # through this manager
            storage.item = _new_boto3_session(region_name=region)
        return cast(boto3.session.Session, storage.item)

    def resource(self, region: Optional[str], service_name: str, endpoint_url: Optional[str] = None) -> boto3.resources.base.ServiceResource:
        """
        Get the Boto3 Resource to use with the given service (like 'ec2') in the given region.

        :param endpoint_url: AWS endpoint URL to use for the client. If not
               specified, a default is used.
        """
        key = (region, service_name, endpoint_url)
        storage = self.resource_cache[key]
        if not hasattr(storage, 'item'):
            with _init_lock:
                # We lock inside the if check; we don't care if the memoization
                # sometimes results in multiple different copies leaking out.
                # We lock because we call .resource()

                if endpoint_url is not None:
                    # The Boto3 stubs are missing an overload for `resource` that takes
                    # a non-literal string. See
                    # <https://github.com/vemel/mypy_boto3_builder/issues/121#issuecomment-1011322636>
                    storage.item = self.session(region).resource(service_name, endpoint_url=endpoint_url) # type: ignore
                else:
                    # We might not be able to pass None to Boto3 and have it be the same as no argument.
                    storage.item = self.session(region).resource(service_name) # type: ignore

        return cast(boto3.resources.base.ServiceResource, storage.item)

    def client(self, region: Optional[str], service_name: str, endpoint_url: Optional[str] = None, config: Optional[Config] = None) -> botocore.client.BaseClient:
        """
        Get the Boto3 Client to use with the given service (like 'ec2') in the given region.

        :param endpoint_url: AWS endpoint URL to use for the client. If not
               specified, a default is used.
        :param config: Custom configuration to use for the client.
        """

        if config is not None:
            # Don't try and memoize if a custom config is used
            with _init_lock:
                if endpoint_url is not None:
                    return self.session(region).client(service_name, endpoint_url=endpoint_url, config=config) # type: ignore
                else:
                    return self.session(region).client(service_name, config=config) # type: ignore

        key = (region, service_name, endpoint_url)
        storage = self.client_cache[key]
        if not hasattr(storage, 'item'):
            with _init_lock:
                # We lock because we call .client()

                if endpoint_url is not None:
                    # The Boto3 stubs are probably missing an overload here too. See:
                    # <https://github.com/vemel/mypy_boto3_builder/issues/121#issuecomment-1011322636>
                    storage.item = self.session(region).client(service_name, endpoint_url=endpoint_url) # type: ignore
                else:
                    # We might not be able to pass None to Boto3 and have it be the same as no argument.
                    storage.item = self.session(region).client(service_name) # type: ignore
        return cast(botocore.client.BaseClient , storage.item)

    def boto2(self, region: Optional[str], service_name: str) -> boto.connection.AWSAuthConnection:
        """
        Get the connected boto2 connection for the given region and service.
        """
        if service_name == 'iam':
            # IAM connections are regionless
            region = 'universal'
        key = (region, service_name)
        storage = self.boto2_cache[key]
        if not hasattr(storage, 'item'):
            with _init_lock:
                storage.item = getattr(boto, service_name).connect_to_region(region, profile_name=os.environ.get("TOIL_AWS_PROFILE", None))
        return cast(boto.connection.AWSAuthConnection, storage.item)

# If you don't want your own AWSConnectionManager, we have a global one and some global functions
_global_manager = AWSConnectionManager()

def establish_boto3_session(region_name: Optional[str] = None) -> Session:
    """
    Get a Boto 3 session usable by the current thread.

    This function may not always establish a *new* session; it can be memoized.
    """

    # Just use a global version of the manager. Note that we change the argument order!
    return _global_manager.session(region_name)

def client(service_name: str, region_name: Optional[str] = None, endpoint_url: Optional[str] = None, config: Optional[Config] = None) -> botocore.client.BaseClient:
    """
    Get a Boto 3 client for a particular AWS service, usable by the current thread.

    Global alternative to AWSConnectionManager.
    """

    # Just use a global version of the manager. Note that we change the argument order!
    return _global_manager.client(region_name, service_name, endpoint_url=endpoint_url, config=config)

def resource(service_name: str, region_name: Optional[str] = None, endpoint_url: Optional[str] = None) -> boto3.resources.base.ServiceResource:
    """
    Get a Boto 3 resource for a particular AWS service, usable by the current thread.

    Global alternative to AWSConnectionManager.
    """

    # Just use a global version of the manager. Note that we change the argument order!
    return _global_manager.resource(region_name, service_name, endpoint_url=endpoint_url)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import shutil
import subprocess
import sys
import time
import uuid
from unittest.mock import patch

import pytest

pkg_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))  # noqa
sys.path.insert(0, pkg_root)  # noqa

import toil
from toil import resolveEntryPoint
from toil.common import Config, Toil
from toil.job import Job
from toil.lib.bioio import system
from toil.test import (ToilTest,
                       get_temp_file,
                       integrative,
                       needs_aws_ec2,
                       needs_cwl,
                       needs_docker,
                       needs_rsync3,
                       slow)
from toil.test.sort.sortTest import makeFileToSort
from toil.utils.toilStats import getStats, processData
from toil.utils.toilStatus import ToilStatus
from toil.version import python

logger = logging.getLogger(__name__)


class UtilsTest(ToilTest):
    """
    Tests the utilities that toil ships with, e.g. stats and status, in conjunction with restart
    functionality.
    """

    def setUp(self):
        super().setUp()
        self.tempDir = self._createTempDir()
        self.tempFile = get_temp_file(rootDir=self.tempDir)
        self.outputFile = 'someSortedStuff.txt'
        self.toilDir = os.path.join(self.tempDir, "jobstore")
        self.assertFalse(os.path.exists(self.toilDir))
        self.lines = 1000
        self.lineLen = 10
        self.N = 1000
        makeFileToSort(self.tempFile, self.lines, self.lineLen)
        # First make our own sorted version
        with open(self.tempFile) as fileHandle:
            self.correctSort = fileHandle.readlines()
            self.correctSort.sort()

        self.sort_workflow_cmd = [
            python,
            '-m',
            'toil.test.sort.sort',
            f'file:{self.toilDir}',
            '--clean=never',
            '--numLines=1',
            '--lineLength=1'
        ]

        self.restart_sort_workflow_cmd = [
            python,
            '-m',
            'toil.test.sort.restart_sort',
            f'file:{self.toilDir}'
        ]

    def tearDown(self):
        if os.path.exists(self.tempDir):
            shutil.rmtree(self.tempDir)
        if os.path.exists(self.toilDir):
            shutil.rmtree(self.toilDir)

        for f in ['fileToSort.txt', 'sortedFile.txt', 'output.txt']:
            if os.path.exists(f):
                os.remove(f)

        ToilTest.tearDown(self)

    @property
    def toilMain(self):
        return resolveEntryPoint('toil')

    @property
    def cleanCommand(self):
        return [self.toilMain, 'clean', self.toilDir]

    @property
    def statsCommand(self):
        return [self.toilMain, 'stats', self.toilDir, '--pretty']

    def statusCommand(self, failIfNotComplete=False):
        commandTokens = [self.toilMain, 'status', self.toilDir]
        if failIfNotComplete:
            commandTokens.append('--failIfNotComplete')
        return commandTokens

    @needs_rsync3
    @pytest.mark.timeout(1200)
    @needs_aws_ec2
    @integrative
    @slow
    def testAWSProvisionerUtils(self):
        """
        Runs a number of the cluster utilities in sequence.

        Launches a cluster with custom tags.
        Verifies the tags exist.
        ssh's into the cluster.
        Does some weird string comparisons.
        Makes certain that TOIL_WORKDIR is set as expected in the ssh'ed cluster.
        Rsyncs a file and verifies it exists on the leader.
        Destroys the cluster.

        :return:
        """
        # TODO: Run these for the other clouds.
        clusterName = f'cluster-utils-test{uuid.uuid4()}'
        keyName = os.getenv('TOIL_AWS_KEYNAME').strip() or 'id_rsa'
        expected_owner = os.getenv('TOIL_OWNER_TAG') or keyName

        try:
            from toil.provisioners.aws.awsProvisioner import AWSProvisioner
            aws_provisioner = AWSProvisioner.__module__
            logger.debug(f"Found AWSProvisioner: {aws_provisioner}.")

            # launch master with an assortment of custom tags
            system([self.toilMain, 'launch-cluster', '--clusterType', 'mesos',
                    '-t', 'key1=value1', '-t', 'key2=value2', '--tag', 'key3=value3',
                    '--leaderNodeType=t2.medium', '--keyPairName=' + keyName, clusterName,
                    '--provisioner=aws', '--zone=us-west-2a', '--logLevel=DEBUG'])

            cluster = toil.provisioners.cluster_factory(provisioner='aws', zone='us-west-2a', clusterName=clusterName)
            leader = cluster.getLeader()

            # check that the leader carries the appropriate tags
            tags = {'key1': 'value1', 'key2': 'value2', 'key3': 'value3', 'Name': clusterName, 'Owner': expected_owner}
            for key in tags:
                self.assertEqual(leader.tags.get(key), tags[key])
        finally:
            system([self.toilMain, 'destroy-cluster', '--zone=us-west-2a', '--provisioner=aws', clusterName])

    @slow
    def testUtilsSort(self):
        """
        Tests the status and stats commands of the toil command line utility using the
        sort example with the --restart flag.
        """
        # Get the sort command to run
        toilCommand = [sys.executable,
                       '-m', toil.test.sort.sort.__name__,
                       self.toilDir,
                       '--logLevel=DEBUG',
                       '--fileToSort', self.tempFile,
                       '--outputFile', self.outputFile,
                       '--N', str(self.N),
                       '--stats',
                       '--retryCount=2',
                       '--badWorker=0.5',
                       '--badWorkerFailInterval=0.05']
        # Try restarting it to check that a JobStoreException is thrown
        self.assertRaises(subprocess.CalledProcessError, system, toilCommand + ['--restart'])
        # Check that trying to run it in restart mode does not create the jobStore
        self.assertFalse(os.path.exists(self.toilDir))

        # Status command
        # Run the script for the first time
        try:
            system(toilCommand)
            finished = True
        except subprocess.CalledProcessError:  # This happens when the script fails due to having unfinished jobs
            system(self.statusCommand())
            self.assertRaises(subprocess.CalledProcessError, system, self.statusCommand(failIfNotComplete=True))
            finished = False
        self.assertTrue(os.path.exists(self.toilDir))

        # Try running it without restart and check an exception is thrown
        self.assertRaises(subprocess.CalledProcessError, system, toilCommand)

        # Now restart it until done
        totalTrys = 1
        while not finished:
            try:
                system(toilCommand + ['--restart'])
                finished = True
            except subprocess.CalledProcessError:  # This happens when the script fails due to having unfinished jobs
                system(self.statusCommand())
                self.assertRaises(subprocess.CalledProcessError, system, self.statusCommand(failIfNotComplete=True))
                if totalTrys > 16:
                    self.fail()  # Exceeded a reasonable number of restarts
                totalTrys += 1

        # Check the toil status command does not issue an exception
        system(self.statusCommand())

        # Check we can run 'toil stats'
        system(self.statsCommand)

        # Check the file is properly sorted
        with open(self.outputFile) as fileHandle:
            l2 = fileHandle.readlines()
            self.assertEqual(self.correctSort, l2)

        # Delete output file before next step
        os.remove(self.outputFile)

        # Check we can run 'toil clean'
        system(self.cleanCommand)

    @slow
    def testUtilsStatsSort(self):
        """
        Tests the stats commands on a complete run of the stats test.
        """
        # Get the sort command to run
        toilCommand = [sys.executable,
                       '-m', toil.test.sort.sort.__name__,
                       self.toilDir,
                       '--logLevel=DEBUG',
                       '--fileToSort', self.tempFile,
                       '--outputFile', self.outputFile,
                       '--N', str(self.N),
                       '--stats',
                       '--retryCount=99',
                       '--badWorker=0.5',
                       '--badWorkerFailInterval=0.01']

        # Run the script for the first time
        system(toilCommand)
        self.assertTrue(os.path.exists(self.toilDir))

        # Check we can run 'toil stats'
        system(self.statsCommand)

        # Check the file is properly sorted
        with open(self.outputFile) as fileHandle:
            l2 = fileHandle.readlines()
            self.assertEqual(self.correctSort, l2)

        # Delete output file
        os.remove(self.outputFile)

    def testUnicodeSupport(self):
        options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
        options.clean = 'always'
        options.logLevel = 'debug'
        Job.Runner.startToil(Job.wrapFn(printUnicodeCharacter), options)

    @slow
    def testMultipleJobsPerWorkerStats(self):
        """
        Tests case where multiple jobs are run on 1 worker to ensure that all jobs report back their data
        """
        options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
        options.clean = 'never'
        options.stats = True
        Job.Runner.startToil(RunTwoJobsPerWorker(), options)
        config = Config()
        config.setOptions(options)
        jobStore = Toil.resumeJobStore(config.jobStore)
        stats = getStats(jobStore)
        collatedStats = processData(jobStore.config, stats)
        self.assertTrue(len(collatedStats.job_types) == 2, "Some jobs are not represented in the stats.")

    def check_status(self, status, status_fn, seconds=20):
        i = 0.0
        while status_fn(self.toilDir) != status:
            time.sleep(0.5)
            i += 0.5
            if i > seconds:
                s = status_fn(self.toilDir)
                self.assertEqual(s, status, f'Waited {seconds} seconds without status reaching {status}; stuck at {s}')

    def testGetPIDStatus(self):
        """Test that ToilStatus.getPIDStatus() behaves as expected."""
        wf = subprocess.Popen(self.sort_workflow_cmd)
        self.check_status('RUNNING', status_fn=ToilStatus.getPIDStatus, seconds=20)
        wf.wait()
        self.check_status('COMPLETED', status_fn=ToilStatus.getPIDStatus)

        # TODO: we need to reach into the FileJobStore's files and delete this
        #  shared file. We assume we know its internal layout.
        os.remove(os.path.join(self.toilDir, 'files/shared/pid.log'))
        self.check_status('QUEUED', status_fn=ToilStatus.getPIDStatus)

    def testGetStatusFailedToilWF(self):
        """
        Test that ToilStatus.getStatus() behaves as expected with a failing Toil workflow.
        While this workflow could be called by importing and evoking its main function, doing so would remove the
        opportunity to test the 'RUNNING' functionality of getStatus().
        """
        # --badWorker is set to force failure.
        wf = subprocess.Popen(self.sort_workflow_cmd + ['--badWorker=1'])
        self.check_status('RUNNING', status_fn=ToilStatus.getStatus)
        wf.wait()
        self.check_status('ERROR', status_fn=ToilStatus.getStatus)

    @needs_cwl
    @needs_docker
    def testGetStatusFailedCWLWF(self):
        """Test that ToilStatus.getStatus() behaves as expected with a failing CWL workflow."""
        # --badWorker is set to force failure.
        cmd = ['toil-cwl-runner', '--jobStore', self.toilDir, '--clean=never', '--badWorker=1',
               'src/toil/test/cwl/sorttool.cwl', '--reverse', '--input', 'src/toil/test/cwl/whale.txt']
        wf = subprocess.Popen(cmd)
        self.check_status('RUNNING', status_fn=ToilStatus.getStatus)
        wf.wait()
        self.check_status('ERROR', status_fn=ToilStatus.getStatus)

    @needs_cwl
    @needs_docker
    def testGetStatusSuccessfulCWLWF(self):
        """Test that ToilStatus.getStatus() behaves as expected with a successful CWL workflow."""
        cmd = ['toil-cwl-runner', '--jobStore', self.toilDir, '--clean=never',
               'src/toil/test/cwl/sorttool.cwl', '--reverse', '--input', 'src/toil/test/cwl/whale.txt']
        wf = subprocess.Popen(cmd)
        self.check_status('RUNNING', status_fn=ToilStatus.getStatus, seconds=20)
        wf.wait()
        self.check_status('COMPLETED', status_fn=ToilStatus.getStatus)

    @needs_cwl
    @patch('builtins.print')
    def testPrintJobLog(self, mock_print):
        """Test that ToilStatus.printJobLog() reads the log from a failed command without error."""
        # Run a workflow that will always fail
        cmd = ['toil-cwl-runner', '--jobStore', self.toilDir, '--clean=never',
               'src/toil/test/cwl/alwaysfails.cwl', '--message', 'Testing']
        wf = subprocess.Popen(cmd)
        wf.wait()
        # print log and check output
        status = ToilStatus(self.toilDir)
        status.printJobLog()

        # Make sure it printed some kind of complaint about the missing command.
        args, kwargs = mock_print.call_args
        self.assertIn('invalidcommand', args[0])

    def testRestartAttribute(self):
        """
        Test that the job store is only destroyed when we observe a succcessful workflow run.
        The following simulates a failing workflow that attempts to resume without restart().
        In this case, the job store should not be destroyed until restart() is called.
        """
        # Run a workflow that will always fail
        cmd = self.restart_sort_workflow_cmd + ['--badWorker=1']
        subprocess.run(cmd)

        restart_cmd = self.restart_sort_workflow_cmd + ['--badWorker=0', '--restart']
        subprocess.run(restart_cmd)

        # Check the job store exists after restart attempt
        self.assertTrue(os.path.exists(self.toilDir))

        successful_cmd = [python, '-m', 'toil.test.sort.sort', 'file:' + self.toilDir,
                                  '--restart']
        subprocess.run(successful_cmd)

        # Check the job store is destroyed after calling restart()
        self.assertFalse(os.path.exists(self.toilDir))


def printUnicodeCharacter():
    # We want to get a unicode character to stdout but we can't print it directly because of
    # Python encoding issues. To work around this we print in a separate Python process. See
    # http://stackoverflow.com/questions/492483/setting-the-correct-encoding-when-piping-stdout-in-python
    subprocess.check_call([sys.executable, '-c', "print('\\xc3\\xbc')"])


class RunTwoJobsPerWorker(Job):
    """
    Runs child job with same resources as self in an attempt to chain the jobs on the same worker
    """
    def __init__(self):
        Job.__init__(self)

    def run(self, fileStore):
        self.addChildFn(printUnicodeCharacter)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import pytest

from toil.lib.aws.ami import (aws_marketplace_flatcar_ami_search,
                              get_flatcar_ami,
                              feed_flatcar_ami_release,
                              flatcar_release_feed_amis)
from toil.lib.aws.session import establish_boto3_session
from toil.test import ToilTest, needs_aws_ec2

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


class FlatcarFeedTest(ToilTest):
    """Test accessing the FLatcar AMI release feed, independent of the AWS API"""
    
    def test_parse_archive_feed(self):
        """Make sure we can get a Flatcar release from the Internet Archive."""
        amis = list(flatcar_release_feed_amis('us-west-2', 'amd64', 'archive'))
        self.assertTrue(len(amis) > 0)
        for ami in amis:
            self.assertEqual(len(ami), len('ami-02b46c73fed689d1c'))
            self.assertTrue(ami.startswith('ami-'))
            
    def test_parse_beta_feed(self):
        """Make sure we can get a Flatcar release from the beta channel."""
        amis = list(flatcar_release_feed_amis('us-west-2', 'amd64', 'beta'))
        self.assertTrue(len(amis) > 0)
        for ami in amis:
            self.assertEqual(len(ami), len('ami-02b46c73fed689d1c'))
            self.assertTrue(ami.startswith('ami-'))
    
    # TODO: This will fail until https://github.com/flatcar/Flatcar/issues/962 is fixed
    @pytest.mark.xfail
    def test_parse_stable_feed(self):
        """Make sure we can get a Flatcar release from the stable channel."""
        amis = list(flatcar_release_feed_amis('us-west-2', 'amd64', 'stable'))
        self.assertTrue(len(amis) > 0)
        for ami in amis:
            self.assertEqual(len(ami), len('ami-02b46c73fed689d1c'))
            self.assertTrue(ami.startswith('ami-'))
            
    def test_bypass_stable_feed(self):
        """Make sure we can either get or safely not get a Flatcar release from the stable channel."""
        list(flatcar_release_feed_amis('us-west-2', 'amd64', 'stable'))
        # Ifd we get here we safely managed to iterate everything.

@needs_aws_ec2
class AMITest(ToilTest):
    @classmethod
    def setUpClass(cls):
        session = establish_boto3_session(region_name='us-west-2')
        cls.ec2_client = session.client('ec2')

    def test_fetch_flatcar(self):
        with self.subTest('Test flatcar AMI from user is prioritized.'):
            os.environ['TOIL_AWS_AMI'] = 'overridden'
            ami = get_flatcar_ami(self.ec2_client)
            self.assertEqual(ami, 'overridden')
            del os.environ['TOIL_AWS_AMI']

        with self.subTest('Test flatcar AMI returns an AMI-looking AMI.'):
            ami = get_flatcar_ami(self.ec2_client)
            self.assertEqual(len(ami), len('ami-02b46c73fed689d1c'))
            self.assertTrue(ami.startswith('ami-'))

        with self.subTest('Test feed_flatcar_ami_release() returns an AMI-looking AMI.'):
            ami = feed_flatcar_ami_release(self.ec2_client, source='archive')
            self.assertTrue(ami is None or len(ami) == len('ami-02b46c73fed689d1c'))
            self.assertTrue(ami is None or ami.startswith('ami-'))

        with self.subTest('Test aws_marketplace_flatcar_ami_search() returns an AMI-looking AMI.'):
            ami = aws_marketplace_flatcar_ami_search(self.ec2_client)
            self.assertEqual(len(ami), len('ami-02b46c73fed689d1c'))
            self.assertTrue(ami.startswith('ami-'))

    # TODO: This will fail until https://github.com/flatcar/Flatcar/issues/962 is fixed
    @pytest.mark.xfail
    def test_fetch_arm_flatcar(self):
        """Test flatcar AMI finder architecture parameter."""
        amis = set()
        for arch in ['amd64', 'arm64']:
            ami = get_flatcar_ami(self.ec2_client, architecture=arch)
            self.assertTrue(ami.startswith('ami-'))
            amis.add(ami)
        self.assertTrue(len(amis) == 2)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import logging
import math
import os
import signal
import subprocess
import time
import traceback
from argparse import ArgumentParser, _ArgumentGroup
from queue import Empty, Queue
from threading import Event, Lock, Thread
from typing import Dict, List, Optional, Set, Sequence, Tuple, Union

import toil
from toil import worker as toil_worker
from toil.batchSystems.abstractBatchSystem import (EXIT_STATUS_UNAVAILABLE_VALUE,
                                                   BatchSystemSupport,
                                                   ResourcePool,
                                                   ResourceSet,
                                                   UpdatedBatchJobInfo,
                                                   InsufficientSystemResources)

from toil.bus import ExternalBatchIdMessage
from toil.batchSystems.options import OptionSetter

from toil.common import SYS_MAX_SIZE, Config, Toil, fC
from toil.job import JobDescription, AcceleratorRequirement, accelerator_satisfies, Requirer
from toil.lib.accelerators import get_individual_local_accelerators, get_restrictive_environment_for_local_accelerators
from toil.lib.threading import cpu_count

logger = logging.getLogger(__name__)


class SingleMachineBatchSystem(BatchSystemSupport):
    """
    The interface for running jobs on a single machine, runs all the jobs you
    give it as they come in, but in parallel.

    Uses a single "daddy" thread to manage a fleet of child processes.

    Communication with the daddy thread happens via two queues: one queue of
    jobs waiting to be run (the input queue), and one queue of jobs that are
    finished/stopped and need to be returned by getUpdatedBatchJob (the output
    queue).

    When the batch system is shut down, the daddy thread is stopped.

    If running in debug-worker mode, jobs are run immediately as they are sent
    to the batch system, in the sending thread, and the daddy thread is not
    run. But the queues are still used.
    """

    @classmethod
    def supportsAutoDeployment(cls):
        return False

    @classmethod
    def supportsWorkerCleanup(cls):
        return True

    numCores = cpu_count()

    minCores = 0.1
    """
    The minimal fractional CPU. Tasks with a smaller core requirement will be rounded up to this
    value.
    """
    physicalMemory = toil.physicalMemory()

    def __init__(
        self, config: Config, maxCores: float, maxMemory: int, maxDisk: int, max_jobs: Optional[int] = None
    ) -> None:
        self.config = config

        if max_jobs is None:
            # Use the main limit for jobs from the config
            self.max_jobs = self.config.max_jobs
        else:
            # Use the override (probably because we are the local batch system
            # and not the main one).
            self.max_jobs = max_jobs

        # Limit to the smaller of the user-imposed limit and what we actually
        # have on this machine for each resource.
        #
        # If we don't have up to the limit of the resource (and the resource
        # isn't the inlimited sentinel), warn.
        if maxCores > self.numCores:
            if maxCores != SYS_MAX_SIZE and maxCores != float('inf'):
                # We have an actually specified limit and not the default
                logger.warning('Not enough cores! User limited to %i but we only have %i.', maxCores, self.numCores)
            maxCores = self.numCores
        if maxMemory > self.physicalMemory:
            if maxMemory != SYS_MAX_SIZE:
                # We have an actually specified limit and not the default
                logger.warning('Not enough memory! User limited to %i bytes but we only have %i bytes.', maxMemory, self.physicalMemory)
            maxMemory = self.physicalMemory

        workdir = Toil.getLocalWorkflowDir(config.workflowID, config.workDir)  # config.workDir may be None; this sets a real directory
        self.physicalDisk = toil.physicalDisk(workdir)
        if maxDisk > self.physicalDisk:
            if maxDisk != SYS_MAX_SIZE:
                # We have an actually specified limit and not the default
                logger.warning('Not enough disk space! User limited to %i bytes but we only have %i bytes.', maxDisk, self.physicalDisk)
            maxDisk = self.physicalDisk

        super().__init__(config, maxCores, maxMemory, maxDisk)
        assert self.maxCores >= self.minCores
        assert self.maxMemory >= 1

        # The scale allows the user to apply a factor to each task's cores requirement, thereby
        # squeezing more tasks onto each core (scale < 1) or stretching tasks over more cores
        # (scale > 1).
        self.scale = config.scale

        if config.badWorker > 0 and config.debugWorker:
            # We can't throw SIGUSR1 at the worker because it is also going to
            # be the leader and/or test harness.
            raise RuntimeError("Cannot use badWorker and debugWorker together; "
                "worker would have to kill the leader")

        self.debugWorker = config.debugWorker

        # A counter to generate job IDs and a lock to guard it.
        # We make sure to start this at 1 so that job IDs are never falsy.
        self.jobIndex = 1
        self.jobIndexLock = Lock()

        # A dictionary mapping batch system IDs of submitted jobs to the command line
        self.jobs: Dict[int, JobDescription] = {}

        # A queue of jobs waiting to be executed. Consumed by the daddy thread.
        self.inputQueue = Queue()

        # A queue of finished jobs. Produced by the daddy thread.
        self.outputQueue = Queue()

        # A dictionary mapping batch system IDs of currently running jobs to their Info objects
        self.runningJobs: Dict[int, Info] = {}

        # These next two are only used outside debug-worker mode

        # A dict mapping PIDs to Popen objects for running jobs.
        # Jobs that don't fork are executed one at a time in the main thread.
        self.children: Dict[int, subprocess.Popen] = {}
        # A dict mapping child PIDs to the Job IDs they are supposed to be running.
        self.childToJob: Dict[int, str] = {}

        # For accelerators, we need a collection of what each accelerator is, and an acquirable set of them.
        self.accelerator_identities = get_individual_local_accelerators()

        # Put them all organized by resource type
        self.resource_sources = [
            # A pool representing available job slots
            ResourcePool(self.max_jobs, 'job slots'),
            # A pool representing available CPU in units of minCores
            ResourcePool(int(self.maxCores / self.minCores), 'cores'),
            # A pool representing available memory in bytes
            ResourcePool(self.maxMemory, 'memory'),
            # A pool representing the available space in bytes
            ResourcePool(self.maxDisk, 'disk'),
            # And a set for acquiring individual accelerators
            ResourceSet(set(range(len(self.accelerator_identities))), 'accelerators')
        ]

        # If we can't schedule something, we fill this in with a reason why
        self.schedulingStatusMessage = None

        # We use this event to signal shutdown
        self.shuttingDown = Event()

        # A thread in charge of managing all our child processes.
        # Also takes care of resource accounting.
        self.daddyThread = None
        # If it breaks it will fill this in
        self.daddyException: Optional[Exception] = None

        if self.debugWorker:
            logger.debug('Started batch system %s in worker debug mode.', id(self))
        else:
            self.daddyThread = Thread(target=self.daddy, daemon=True)
            self.daddyThread.start()
            logger.debug('Started batch system %s in normal mode.', id(self))

    def daddy(self):
        """
        Be the "daddy" thread.

        Our job is to look at jobs from the input queue.

        If a job fits in the available resources, we allocate resources for it
        and kick off a child process.

        We also check on our children.

        When a child finishes, we reap it, release its resources, and put its
        information in the output queue.
        """

        try:
            logger.debug('Started daddy thread for batch system %s.', id(self))

            while not self.shuttingDown.is_set():
                # Main loop

                while not self.shuttingDown.is_set():
                    # Try to start as many jobs as we can try to start
                    try:
                        # Grab something from the input queue if available.
                        args = self.inputQueue.get_nowait()
                        jobCommand, jobID, jobCores, jobMemory, jobDisk, job_accelerators, environment = args

                        coreFractions = int(jobCores / self.minCores)

                        # Try to start the child
                        result = self._startChild(jobCommand, jobID,
                            coreFractions, jobMemory, jobDisk, job_accelerators, environment)

                        if result is None:
                            # We did not get the resources to run this job.
                            # Requeue last, so we can look at the next job.
                            # TODO: Have some kind of condition the job can wait on,
                            # but without threads (queues for jobs needing
                            # cores/memory/disk individually)?
                            self.inputQueue.put(args)
                            break
                        elif result is not False:
                            #Result is a PID

                            if self._outbox is not None:
                                # Annotate the job with the PID generated.
                                self._outbox.publish(
                                   ExternalBatchIdMessage(jobID, str(result), self.__class__.__name__))

                        # Otherwise False

                    except Empty:
                        # Nothing to run. Stop looking in the queue.
                        break

                # Now check on our children.
                for done_pid in self._pollForDoneChildrenIn(self.children):
                    # A child has actually finished.
                    # Clean up after it.
                    self._handleChild(done_pid)

                # Then loop again: start and collect more jobs.
                # TODO: It would be good to be able to wait on a new job or a finished child, whichever comes first.
                # For now we just sleep and loop.
                time.sleep(0.01)


            # When we get here, we are shutting down.
            logger.debug('Daddy thread cleaning up %d remaining children for batch system %s...', len(self.children), id(self))

            self._stop_and_wait(self.children.values())

            logger.debug('Daddy thread for batch system %s finishing because no children should now exist', id(self))

            # Then exit the thread.
            return
        except Exception as e:
            logger.critical('Unhandled exception in daddy thread for batch system %s: %s', id(self), traceback.format_exc())
            # Pass the exception back to the main thread so it can stop the next person who calls into us.
            self.daddyException = e
            raise

    def _checkOnDaddy(self):
        if self.daddyException is not None:
            # The daddy thread broke and we cannot do our job
            logger.critical('Propagating unhandled exception in daddy thread to main thread')
            exc = self.daddyException
            self.daddyException = None
            if isinstance(exc, Exception):
                raise exc
            else:
                raise TypeError(f'Daddy thread failed with non-exception: {exc}')

    def _stop_now(self, popens: Sequence[subprocess.Popen]) -> List[int]:
        """
        Stop the given child processes and all their children. Does not reap them.

        Returns a list of PGIDs killed, where processes may exist that have not
        yet received their kill signals.
        """

        # We will potentially need to poll these PGIDs to ensure that all
        # processes in them are gone.
        pgids = []

        for popen in popens:
            # Kill all the children

            if popen.returncode is None:
                # Process is not known to be dead. Try and grab its group.
                try:
                    pgid = os.getpgid(popen.pid)
                except OSError:
                    # It just died. Assume the pgid was its PID.
                    pgid = popen.pid
            else:
                # It is dead. Try it's PID as a PGID and hope we didn't re-use it.
                pgid = popen.pid

            if pgid != os.getpgrp():
                # The child process really is in its own group, and not ours.

                # Kill the group, which hopefully hasn't been reused
                logger.debug('Send shutdown kill to process group %s known to batch system %s', pgid, id(self))
                try:
                    os.killpg(pgid, signal.SIGKILL)
                    pgids.append(pgid)
                except ProcessLookupError:
                    # It is dead already
                    pass
                except PermissionError:
                    # It isn't ours actually. Ours is dead.
                    pass
            else:
                # Kill the subprocess again through popen in case it somehow
                # never managed to make the group.
                popen.kill()

        return pgids

    def _stop_and_wait(self, popens: Sequence[subprocess.Popen], timeout: int = 5) -> None:
        """
        Stop the given child processes and all their children. Blocks until the
        processes are gone or timeout is passed.

        :param popens: The processes to stop and wait on.
        :param timeout: The number of seconds to wait for all process groups to
                        be gone.
        """

        pgids = self._stop_now(popens)

        for popen in popens:
            # Wait on all the children
            popen.wait()
            logger.debug('Process %s known to batch system %s is stopped; it returned %s',
                         popen.pid, id(self), popen.returncode)

        # Make sure all child processes have received their kill signal
        self._wait_for_death(pgids, timeout)

    def _wait_for_death(self, pgids: List[int], timeout: int = 5):
        """
        Wait for the process groups to be killed. Blocks until the processes
        are gone or timeout is passed.

        :param pgids: The list of process group ids.
        :param timeout: The number of seconds to wait for all process groups to
                        be gone.
        """
        # TODO: this opens a PGID reuse risk; someone else might've reaped the
        #  process and its PGID may have been re-used.

        start = datetime.datetime.now()
        while len(pgids) > 0 and (datetime.datetime.now() - start).total_seconds() < timeout:
            new_pgids: List[int] = []
            for pgid in pgids:
                try:
                    # Send a kill to the group again, to see if anything in it
                    # is still alive. Our first kill might not have been
                    # delivered yet.
                    os.killpg(pgid, signal.SIGKILL)

                    # If we reach here, something in the process group still
                    # exists.
                    new_pgids.append(pgid)
                except ProcessLookupError:
                    # The group is actually gone now.
                    pass
                except PermissionError:
                    # The group is not only gone but reused
                    pass

            pgids = new_pgids
            if len(pgids) > 0:
                time.sleep(0.1)

        if len(pgids) > 0:
            # If any processes are still alive, let user know that we may leave
            # behind dead but unreaped processes.
            logger.warning('Processes were not reaped in groups: %s.', str(pgids))
            logger.warning('Make sure your jobs are cleaning up child processes appropriately to avoid zombie '
                           'processes possibly being left behind.')

    def _pollForDoneChildrenIn(self, pid_to_popen):
        """
        See if any children represented in the given dict from PID to Popen
        object have finished.

        Return a collection of their PIDs.

        Guarantees that each child's exit code will be gettable via wait() on
        the child's Popen object (i.e. does not reap the child, unless via
        Popen).
        """

        # We keep our found PIDs in a set so we can work around waitid showing
        # us the same one repeatedly.
        ready = set()

        # Find the waitid function
        waitid = getattr(os, 'waitid', None)

        if callable(waitid):
            # waitid exists (not Mac)

            while True:
                # Poll for any child to have exit, but don't reap it. Leave reaping
                # to the Popen.
                # TODO: What if someone else in Toil wants to do this syscall?
                # TODO: Is this one-notification-per-done-child with WNOHANG? Or
                # can we miss some? Or do we see the same one repeatedly until it
                # is reaped?
                try:
                    siginfo = waitid(os.P_ALL, -1, os.WEXITED | os.WNOWAIT | os.WNOHANG)
                except ChildProcessError:
                    # This happens when there is nothing to wait on right now,
                    # instead of the weird C behavior of overwriting a field in
                    # a pointed-to struct.
                    siginfo = None
                if siginfo is not None and siginfo.si_pid in pid_to_popen and siginfo.si_pid not in ready:
                    # Something new finished
                    ready.add(siginfo.si_pid)
                else:
                    # Nothing we own that we haven't seen before has finished.
                    return ready
        else:
            # On Mac there's no waitid and no way to wait and not reap.
            # Fall back on polling all the Popen objects.
            # To make this vaguely efficient we have to return done children in
            # batches.
            for pid, popen in pid_to_popen.items():
                if popen.poll() is not None:
                    # Process is done
                    ready.add(pid)
                    logger.debug('Child %d has stopped', pid)

            # Return all the done processes we found
            return ready

    def _runDebugJob(self, jobCommand, jobID, environment):
        """
        Run the jobCommand right now, in the current thread.
        May only be called in debug-worker mode.
        Assumes resources are available.
        """
        assert self.debugWorker
        # TODO: It is not possible to kill running jobs in forkless mode,
        # because they are run immediately in the main thread.
        info = Info(time.time(), None, None, killIntended=False)
        self.runningJobs[jobID] = info

        if jobCommand.startswith("_toil_worker "):
            # We can actually run in this thread
            jobName, jobStoreLocator, jobStoreID = jobCommand.split()[1:4] # Parse command
            jobStore = Toil.resumeJobStore(jobStoreLocator)
            toil_worker.workerScript(jobStore, jobStore.config, jobName, jobStoreID,
                                     redirectOutputToLogFile=not self.debugWorker) # Call the worker
        else:
            # Run synchronously. If starting or running the command fails, let the exception stop us.
            subprocess.check_call(jobCommand,
                                  shell=True,
                                  env=dict(os.environ, **environment))

        self.runningJobs.pop(jobID)
        if not info.killIntended:
            self.outputQueue.put(UpdatedBatchJobInfo(jobID=jobID, exitStatus=0, wallTime=time.time() - info.time, exitReason=None))

    def getSchedulingStatusMessage(self):
        # Implement the abstractBatchSystem's scheduling status message API
        return self.schedulingStatusMessage

    def _setSchedulingStatusMessage(self, message):
        """
        If we can't run a job, we record a short message about why not. If the
        leader wants to know what is up with us (for example, to diagnose a
        deadlock), it can ask us for the message.
        """

        self.schedulingStatusMessage = message

    def check_resource_request(self, requirer: Requirer) -> None:
        try:
            super().check_resource_request(requirer)
        except InsufficientSystemResources as e:
            # Tack the scale onto the exception
            e.details.append(f'Scale is set to {self.scale}.')
            raise e

    def _check_accelerator_request(self, requirer: Requirer) -> None:
        _, problem = self._identify_sufficient_accelerators(requirer.accelerators, set(range(len(self.accelerator_identities))))
        if problem is not None:
            # We can't get the accelerators
            raise InsufficientSystemResources(requirer, 'accelerators', self.accelerator_identities, details=[
                f'The accelerator {problem} could not be provided.'
            ])


    def _release_acquired_resources(self, resources: List[Union[int, Set[int]]]) -> None:
        """
        Release all resources acquired for a job.
        Assumes resources are in the order: core fractions, memory, disk, accelerators.
        """

        # What pools and sets do we want resources from

        for resource, request in zip(self.resource_sources, resources):
            assert ((isinstance(resource, ResourcePool) and isinstance(request, int)) or
                    (isinstance(resource, ResourceSet) and isinstance(request, set)))
            resource.release(request)

    def _identify_sufficient_accelerators(self, needed_accelerators: List[AcceleratorRequirement], available_accelerator_ids: Set[int]) -> Tuple[Optional[Set[int]], Optional[AcceleratorRequirement]]:
        """
        Given the accelerator requirements of a job, and the set of available
        accelerators out of our associated collection of accelerators, find a
        set of the available accelerators that satisfies the job's
        requirements.

        Returns that set and None if the set exists, or None and an unsatisfied
        AcceleratorRequirement if it does not.

        TODO: Uses a simple greedy algorithm and not a smart matching
        algorithm, so if the job requires different kinds of accelerators, and
        some accelerators available can match multiple requirements, then it is
        possible that a solution will not be found.

        Ignores accelerator model constraints.
        """
        accelerators_needed: Set[int] = set()
        accelerators_still_available = set(available_accelerator_ids)
        for requirement in needed_accelerators:
            for i in range(requirement['count']):
                # For each individual accelerator we need
                satisfied = False
                for candidate_index in accelerators_still_available:
                    # Check all the ones we haven't grabbed yet
                    # TODO: We'll re-check early ones against this requirement if it has a count of more than one.
                    candidate = self.accelerator_identities[candidate_index]
                    if accelerator_satisfies(candidate, requirement, ignore=['model']):
                        # If this accelerator can satisfy one unit of this requirement.
                        # We ignore model constraints because as a single
                        # machine we can't really determine the models of
                        # installed accelerators in a way consistent with how
                        # workflows are going to name them, and usually people
                        # use one model anyway.

                        # Say we want it
                        accelerators_needed.add(candidate_index)
                        accelerators_still_available.remove(candidate_index)
                        # And move on to the next required unit
                        satisfied = True
                        break
                if not satisfied:
                    # We can't get the resources we need to run right now.
                    return None, requirement
        # If we get here we satisfied everything
        return accelerators_needed, None

    def _startChild(self, jobCommand, jobID, coreFractions, jobMemory, jobDisk, job_accelerators: List[AcceleratorRequirement], environment):
        """
        Start a child process for the given job.

        Allocate its required resources and save it and save it in our bookkeeping structures.

        If the job is started, returns its PID.
        If the job fails to start, reports it as failed and returns False.
        If the job cannot get the resources it needs to start, returns None.
        """

        # We fill this in if we manage to actually start the child.
        popen = None

        # This is when we started working on the job.
        startTime = time.time()

        # And what do we want from each resource in self.resource_sources?
        # We know they go job slot, cores, memory, disk, accelerators.
        resource_requests: List[Union[int, Set[int]]] = [1, coreFractions, jobMemory, jobDisk]

        # Keep a reference to the accelerators separately
        accelerators_needed = None

        if job_accelerators:
            # Try and find some accelerators to use.
            # Start with all the accelerators that are free right now
            accelerator_set : ResourceSet = self.resource_sources[3]
            snapshot = accelerator_set.get_free_snapshot()
            # And build a plan of the ones we want
            accelerators_needed, problem = self._identify_sufficient_accelerators(job_accelerators, snapshot)
            if accelerators_needed is not None:
                # Now we have a plan to get the accelerators we need.
                resource_requests.append(accelerators_needed)
            else:
                # We couldn't make a plan; the accelerators are busy
                assert problem is not None
                logger.debug('Accelerators are busy: %s', problem)
                self._setSchedulingStatusMessage('Not enough accelerators to run job %s' % jobID)
                return None


        acquired = []
        for source, request in zip(self.resource_sources, resource_requests):
            # For each kind of resource we want, go get it
            assert ((isinstance(source, ResourcePool) and isinstance(request, int)) or
                    (isinstance(source, ResourceSet) and isinstance(request, set)))
            if source.acquireNow(request):
                acquired.append(request)
            else:
                # We can't get everything
                self._setSchedulingStatusMessage('Not enough {} to run job {}'.format(source.resource_type, jobID))
                self._release_acquired_resources(acquired)
                return None

        # Now we have all the resources!

        # Prepare the environment
        child_environment = dict(os.environ, **environment)

        # Communicate the accelerator resources, if any, to the child process
        # by modifying the environemnt
        accelerators_acquired: Set[int] = accelerators_needed if accelerators_needed is not None else set()
        child_environment.update(get_restrictive_environment_for_local_accelerators(accelerators_acquired))

        # Actually run the job.
        # When it finishes we will release what it was using.
        # So it is important to not lose track of the child process.

        try:
            # Launch the job.
            # Make sure it is in its own session (and thus its own
            # process group) so that, if the user signals the
            # workflow, Toil will be responsible for killing the
            # job. This also makes sure that we can signal the job
            # and all its children together. We assume that the
            # process group ID will equal the PID of the process we
            # are starting.
            popen = subprocess.Popen(jobCommand,
                                     shell=True,
                                     env=child_environment,
                                     start_new_session=True)
        except Exception:
            # If the job can't start, make sure we release resources now
            self._release_acquired_resources(acquired)

            logger.error('Could not start job %s: %s', jobID, traceback.format_exc())

            # Report as failed.
            self.outputQueue.put(UpdatedBatchJobInfo(jobID=jobID, exitStatus=EXIT_STATUS_UNAVAILABLE_VALUE, wallTime=0, exitReason=None))

            # Complain it broke.
            return False
        else:
            # If the job did start, record it
            self.children[popen.pid] = popen
            # Make sure we can look it up by PID later
            self.childToJob[popen.pid] = jobID
            # Record that the job is running, and the resources it is using
            info = Info(startTime, popen, acquired, killIntended=False)
            self.runningJobs[jobID] = info

            logger.debug('Launched job %s as child %d', jobID, popen.pid)

            # Report success starting the job
            # Note that if a PID were somehow 0 it would look like False
            assert popen.pid != 0
            return popen.pid

    def _handleChild(self, pid: int) -> None:
        """
        Handle a child process PID that has finished.
        The PID must be for a child job we started.
        Not thread safe to run at the same time as we are making more children.

        Remove the child from our bookkeeping structures and free its resources.
        """

        # Look up the child
        popen = self.children[pid]
        jobID = self.childToJob[pid]
        info = self.runningJobs[jobID]

        # Get the job resources reserved by the job
        acquired = info.resources


        # Clean up our records of the job.
        self.runningJobs.pop(jobID)
        self.childToJob.pop(pid)
        self.children.pop(pid)

        if popen.returncode is None or not callable(getattr(os, 'waitid', None)):
            # It isn't reaped yet, or we have to reap all children to see if thay're done.
            # Before we reap it (if possible), kill its PID as a PGID to make sure
            # it isn't leaving children behind.
            # TODO: This is a PGID re-use risk on Mac because the process is
            #  reaped already and the PGID may have been reused.
            try:
                os.killpg(pid, signal.SIGKILL)
            except ProcessLookupError:
                # It is dead already
                pass
            except PermissionError:
                # It isn't ours actually. Ours is dead.
                pass

        # See how the child did, and reap it.
        statusCode = popen.wait()
        if statusCode != 0 and not info.killIntended:
            logger.error("Got exit code %i (indicating failure) "
                      "from job %s.", statusCode, self.jobs[jobID])
        if not info.killIntended:
            # Report if the job failed and we didn't kill it.
            # If we killed it then it shouldn't show up in the queue.
            self.outputQueue.put(UpdatedBatchJobInfo(jobID=jobID, exitStatus=statusCode, wallTime=time.time() - info.time, exitReason=None))

        # Last attempt to make sure all processes in the group have received
        # their kill signals.
        self._wait_for_death([pid])

        # Free up the job's resources.
        self._release_acquired_resources(acquired)

        logger.debug('Child %d for job %s succeeded', pid, jobID)

    def issueBatchJob(self, jobDesc: JobDescription, job_environment: Optional[Dict[str, str]] = None) -> int:
        """Adds the command and resources to a queue to be run."""

        self._checkOnDaddy()

        # Apply scale in cores
        scaled_desc = jobDesc.scale('cores', self.scale)
        # Round cores up to multiples of minCores
        scaled_desc.cores = max(math.ceil(scaled_desc.cores / self.minCores) * self.minCores, self.minCores)

        # Don't do our own assertions about job size vs. our configured size.
        # The abstract batch system can handle it.
        self.check_resource_request(scaled_desc)
        logger.debug(f"Issuing the command: {jobDesc.command} with {scaled_desc.requirements_string()}")
        with self.jobIndexLock:
            jobID = self.jobIndex
            self.jobIndex += 1
        self.jobs[jobID] = jobDesc.command

        environment = self.environment.copy()
        if job_environment:
            environment.update(job_environment)

        if self.debugWorker:
            # Run immediately, blocking for return.
            # Ignore resource requirements; we run one job at a time
            self._runDebugJob(jobDesc.command, jobID, environment)
        else:
            # Queue the job for later
            self.inputQueue.put((jobDesc.command, jobID, scaled_desc.cores, scaled_desc.memory,
                                scaled_desc.disk, scaled_desc.accelerators, environment))

        return jobID

    def killBatchJobs(self, jobIDs: List[int]) -> None:
        """Kills jobs by ID."""

        self._checkOnDaddy()

        logger.debug(f'Killing jobs: {jobIDs}')

        # Collect the popen handles for the jobs we have to stop
        popens: List[subprocess.Popen] = []

        for jobID in jobIDs:
            if jobID in self.runningJobs:
                info = self.runningJobs[jobID]
                info.killIntended = True
                if info.popen is not None:
                    popens.append(info.popen)
                else:
                    # No popen if running in forkless mode currently
                    assert self.debugWorker
                    logger.critical("Can't kill job: %s in debug mode" % jobID)

        # Stop them all in a batch. Don't reap, because we need the daddy
        # thread to reap them to mark the jobs as not running anymore.
        self._stop_now(popens)

        for jobID in jobIDs:
            while jobID in self.runningJobs:
                # Wait for the daddy thread to collect them.
                time.sleep(0.01)

    def getIssuedBatchJobIDs(self) -> List[int]:
        """Just returns all the jobs that have been run, but not yet returned as updated."""

        self._checkOnDaddy()

        return list(self.jobs.keys())

    def getRunningBatchJobIDs(self) -> Dict[int, float]:

        self._checkOnDaddy()

        now = time.time()
        return {jobID: now - info.time for jobID, info in list(self.runningJobs.items())}

    def shutdown(self) -> None:
        """Terminate cleanly and join daddy thread."""
        if self.daddyThread is not None:
            # Tell the daddy thread to stop.
            self.shuttingDown.set()
            # Wait for it to stop.
            self.daddyThread.join()

        BatchSystemSupport.workerCleanup(self.workerCleanupInfo)

    def getUpdatedBatchJob(self, maxWait: int) -> Optional[UpdatedBatchJobInfo]:
        """Returns a tuple of a no-longer-running job, the return value of its process, and its runtime, or None."""

        self._checkOnDaddy()

        try:
            item = self.outputQueue.get(timeout=maxWait)
        except Empty:
            return None
        self.jobs.pop(item.jobID)
        logger.debug("Ran jobID: %s with exit value: %i", item.jobID, item.exitStatus)
        return item

    @classmethod
    def add_options(cls, parser: Union[ArgumentParser, _ArgumentGroup]) -> None:
        parser.add_argument("--scale", dest="scale", default=1,
                            help="A scaling factor to change the value of all submitted tasks's submitted cores.  "
                                 "Used in the single_machine batch system. Useful for running workflows on "
                                 "smaller machines than they were designed for, by setting a value less than 1. "
                                 "(default: %(default)s)")

    @classmethod
    def setOptions(cls, setOption: OptionSetter):
        setOption("scale", float, fC(0.0), default=1)


class Info:
    """
    Record for a running job.

    Stores the start time of the job, the Popen object representing its child
    (or None), the tuple of (coreFractions, memory, disk) it is using (or
    None), and whether the job is supposed to be being killed.
    """
    # Can't use namedtuple here since killIntended needs to be mutable
    def __init__(self, startTime, popen, resources, killIntended):
        self.time = startTime
        self.popen = popen
        self.resources = resources
        self.killIntended = killIntended
